#!/usr/bin/env python
# -*- coding: utf-8 -*-
#--------------------------------------------------------------------
# Author: Mazhugin Aleksey
# Created: 2007/10/01
# ID: $Id: dbg.py 13 2008-11-29 18:31:31Z alex $
# URL: $URL: file:///myfiles/svn_/XPyLIB/trunc/XPyLIB/dbg.py $
# Copyright: Copyright (c) 2007, Mazhugin Aleksey
# License: BSD
#--------------------------------------------------------------------

"""
Simple debug module.

Used for fast debugging. You may set debug output into two different
streams: L{outstream} and L{fstream}. Preferred assignment for streams is:
    - outstream - any console output (by default I{stderr}.
    - fstream - any file output (default is "dbg.log" file into "~/.XPyLIB/logs"),
        where "~" is a user home directory. A dir mode is 0660. If path is
        not exist then created automaticaly.
Function for file output starts with B{f} letter.
Initialization of streams (open file) is doing only if calling writing
function and stream is not initialized. Before start debug output you can
initialize streams you own objects which support B{write} method.
Then program exit autocalling L{shutdown} function which close L{fstream}
if it was opened. If in program no debug output then streams are not
created and file are not opened.

Also you may use indentation level by calling L{lvlup}(), L{lvldown}() and
decorators L{level}, L{wlevel} and L{fwlevel}.
Indentation level used only in next functions: L{fwriteln}()
and L{writeln}(). For tune indentation use L{levelbodychar},
L{levelendchar}, L{levelwidth}.


Debug output example, see L{_debug})::
    
    *** DEBUG START ***
    
    >456
    ..>tst1:test
    >CALL tst2()
    ..>tst2 start
    ..>tst1:from tst2
    ..>CALL tst3()
    ....>tst3
    ..>RET from tst3()
    ..>tst2 end
    >RET from tst2()
    >CALL tst2()
    ..>tst2 start
    ..>tst1:from tst2
    ..>CALL tst3()
    ....>tst3
    ..>RET from tst3()
    ..>tst2 end
    >RET from tst2()
    
    
    *** DEBUG END ***


@var outstream: Output stream, default is I{sys.stdout}.
@type outstream: stream(file-like object).

@var fstream: Output stream, default is "dbg.log" file into "~/.XPyLIB/logs"
    directory or sys.stderr if file is not available. This stream autoclosed
    then exit from application (used atexit module).
@type fstream: stream(file-like object).
@var levelbodychar: Character that fill indentation spaces.
@type levelbodychar: str
@var levelendchar: Character that insert before string output but after
    level indentation.
@type levelendchar: str
@var levelwidth: Number of indentation characters in one indentation level.
@type levelwidth: int
@var levelerr: Error message then level indentation is invalid thrn return from
    wrapped function.
@type levelerr: int
@var _level: Indentation level.
@type _level: int

@var descfuncall: Function call description.
@type descfuncall: str
@var descfunret: Function return description.
@type descfunret: str

@var desceval: String that prepend to eval expression.
@type desceval: str
@var descevalrez: String that prepend to eval resultat.
@type descevalrez: str
@var descevalexc: String that prepend to eval exception.
@type descevalexc: str
@var descevalln: String that added at end of eval output.
@type descevalln: str
"""

import sys
import atexit
import os
from functools import wraps

outstream = None
fstream = None

levelbodychar = '.'
levelendchar = '>'
levelwidth = 2
levelerr = '*** LEVEL ERROR = '
descfuncall = 'CALL '
descfunret = 'RET from '
desceval = 'EVAL: '
descevalrez = '  ER='
descevalexc = '  EXCEPTION:'
descevalln = ' '

_level = 0

def lvlup():
    """
    Increase indentation level.
    """
    global _level
    _level+=1
    
def lvldown():
    """
    Decrease indentation level.
    """
    global _level
    _level-=1
    if _level < 0:
        _level = 0
        
def lvlreset():
    """
    Reset indentation level to "0".
    """
    global _level
    _level = 0
    
def level(fun):
    """
    Auto level decorator for functions.
    
    Call L{lvlup}() befor function call and L{lvldown}() after 
    wrapped function return.
    @param fun: Wrapped function.
    @type fun: func
    """
    @wraps(fun)
    def wrapper(*args, **kwds):
        global _level
        l=_level
        lvlup()
        exc=False
        try:
            r = fun(*args, **kwds)
        except Exception, e:
            exc=e
        ##lvldown()
        _level=l # Defense from error
        if exc:
            raise exc
        return r
    return wrapper

def fwlevel(fun):
    """
    Auto level and name decorator for functions.
    
    Write first 'Call "fun.__name__"' using L{fwriteln}().
    Also as L{level}().
    If level error tested then write '*** DBG LEVEL ERROR = different'
    where 'different' is level different.
    Write last 'Return from "fun.__name__"' using L{fwriteln}().
    @param fun: Wrapped function.
    @type fun: func
    """
    @wraps(fun)
    def wrapper(*args, **kwds):
        global _level
        fwriteln(descfuncall, fun.__name__, '()')
        l=_level
        lvlup()
        exc=False
        try:
            r = fun(*args, **kwds)
        except Exception, e:
            exc=e
        lvldown()
        if l != _level:
            d=_level-l
            _level=l
            fwriteln(levelerr, d)
        fwriteln(descfunret, fun.__name__, '()')
        if exc:
            raise exc
        return r
    return wrapper
    
def wlevel(fun):
    """
    Auto level and name decorator for functions.
    
    Write first 'Call "fun.__name__"' using L{writeln}().
    Also as L{level}().
    If level error tested then write '*** DBG LEVEL ERROR = different'
    where 'different' is level different.
    Write last 'Return from "fun.__name__"' using L{writeln}().
    @param fun: Wrapped function.
    @type fun: func
    """
    @wraps(fun)
    def wrapper(*args, **kwds):
        global _level
        writeln(descfuncall, fun.__name__, '()')
        l=_level
        lvlup()
        exc=False
        try:
            r = fun(*args, **kwds)
        except Exception, e:
            exc=e
        lvldown()
        if l != _level:
            d=_level-l
            _level=l
            fwriteln(levelerr, d)
        writeln(descfunret, fun.__name__, '()')
        if exc:
            raise exc
        return r
    return wrapper
        
def shutdown():
    """
    Close opened file stream L{fstream}.
    """
    global fstream
    global outstream
    if fstream:
        fstream.write('\n\n*** DEBUG END ***')
        if fstream is not sys.stderr:
            fstream.close()
        fstream=None
    if outstream:
        outstream.write('\n\n*** DEBUG END ***')
    
def outstream_init():
    """
    Initialize output stream.
    """
    global outstream
    if not outstream:
        outstream = sys.stderr
        outstream.write('*** DEBUG START ***\n\n')
  
def fstream_init():
    """
    Initialize file output stream.
    
    Default is "dbg.log" file into "~/.XPyLIB/logs"
    directory or sys.stderr if file is not available.
    Where "~" is a user home directory. Path autocreate if need.
    Mode of dirs is 0660.
    """
    global fstream
    if not fstream:
        try:
            p=os.path.expanduser("~")
            p=os.path.join(p, ".xpylib", "logs")
            if not os.path.exists(p):
                os.makedirs(p,0660) ## rw-rw----
            p=os.path.join(p,'dbg.log')
            fstream=open(p,'w',0)
            
##            p=os.path.dirname(os.path.abspath(__file__))
##            if p.endswith('.zip'):
##                p=os.path.dirname(p)
##            p=os.path.join(os.path.dirname(p), 'logs','dbg.log')
##            fstream=open(p,'w',0)
            
        except Exception, e:
            fstream=sys.stderr
            fstream.write('Can\'t open file for debug logging.\nerror:' + \
                str(e) + '\n\n')  
        fstream.write('*** DEBUG START ***\n\n')

        
def fwrite(*arg):
    """
    Write concatenated arguments into outstream.
    
    @param arg: Object to output. All objects convert to string I{str()} function.
    @type arg:  *object
    @return: None.
    @rtype: None
    """
    fstream_init()
    fstream.write( reduce(lambda a, b: a + str(b), arg, '') )

    
def write(*arg):
    """
    Write concatenated arguments into outstream.
    
    @param arg: Object to output. All objects convert to string I{str()} function.
    @type arg:  *object
    @return: None.
    @rtype: None
    """
    outstream_init()
    outstream.write( reduce(lambda a, b: a + str(b), arg, '') )
    
def _writeln(fun, *arg):
    """
    Write concatenated arguments into outstream and append new line ('\\n'). For
    output used fun.
    
    @param fun: Function for output (L{write} or L{fwrite} for example).
    @type fun: func
    @param arg: Object to output. All objects convert to string I{str()} function.
    @type arg:  *object
    @return: None.
    @rtype: None
    @note: For work this function will be call L{fwrite} function.
    """
    fun((levelbodychar * levelwidth * _level) + levelendchar)
    fun(*arg )
    fun('\n' )

def fwriteln(*arg):
    """
    Write concatenated arguments into fstream and append new line ('\\n').
    
    @param arg: Object to output. All objects convert to string I{str()} function.
    @type arg:  *object
    @return: None.
    @rtype: None
    @note: For work this function will be used L{fwrite} function.
    """
    _writeln(fwrite, *arg)
    
def writeln(*arg):
    """
    Write concatenated arguments into outstream and append new line ('\\n').
    
    @param arg: Object to output. All objects convert to string I{str()} function.
    @type arg:  *object
    @return: None.
    @rtype: None
    @note: For work this function will be used L{write} function.
    """
    _writeln(write, *arg)
        
def _rwrite(fun, *arg):
    """
    As L{fwrite} but return arguments tuple and for output use fun.
    
    @param fun: Function for output (L{write} or L{fwrite} for example).
    @type fun: func
    @param arg: Object to output. All objects convert to string I{str()} function.
    @type arg:  *object
    @return: Arguments tuple.
    @rtype: tuple
    """
    fun(*arg)
    return arg

def frwrite(*arg):
    """
    As L{fwrite} but return arguments tuple.
    
    @param arg: Object to output. All objects convert to string I{str()} function.
    @type arg:  *object
    @return: Arguments tuple.
    @rtype: tuple
    """
    return _rwrite(fwrite, *arg)
    
def rwrite(*arg):
    """
    As L{write} but return arguments tuple.
    
    @param arg: Object to output. All objects convert to string I{str()} function.
    @type arg:  *object
    @return: Arguments tuple.
    @rtype: tuple
    """
    return _rwrite(write, *arg)

def _eval(fun, expr, g=globals(), l=locals()):
    """
    Write expr, evaluate it and write result to output using fun.
    
    Write algorithm:
        - write L{desceval} and expr in quotes.
        - evaluate expr using eval.
        - if exception occur then write L{descevalexc} and quoted exception.
        - if no error then write L{descevalrez} and quoted result.
        - write L{descevalln}.
    
    Then write this function you must get globals() and locals() if you using
    global or local scope variables into exprwssion.
    
    For example see L{weval}.
    
    @param fun: Function for output (L{write} or L{fwrite} for example).
    @type fun: func
    @param expr: Expression string.
    @type expr: str
    @param g: Global variables.
    @type g: dict = globals()
    @param l: Local variables.
    @type l: dict = locals()
    @return: Result of evaluation or Exception.
    @rtype: object or Exception
    """
    fun(desceval, '"', expr, '"')
    exc = False
    try:
        er = eval(expr, g, l)
    except Exception, e:
        exc = True
        er = e
    if exc:
        fun(descevalexc, '"', er, '"')
    else:
        fun(descevalrez, '"', er, '"')
    if descevalln:
        fun(descevalln)
    return er
        
def weval(expr, g=globals(), l=locals()):
    """
    Write expr, evaluate it and write result to outstream.
    
    Write algorithm:
        - write L{desceval} and expr in quotes.
        - evaluate expr using eval.
        - if exception occur then write L{descevalexc} and quoted exception.
        - if no error then write L{descevalrez} and quoted result.
        - write L{descevalln}.
        
    Then write this function you must get globals() and locals() if you using
    global or local scope variables into exprwssion.
    
    For example: C{ dbg.weval('dir(a)', globals(), locals()) }
    
    @param expr: Expression string.
    @type expr: str
    @param g: Global variables.
    @type g: dict = globals()
    @param l: Local variables.
    @type l: dict = locals()
    @return: Result of evaluation or Exception.
    @rtype: object or Exception
    """
    return _eval(writeln, expr,g, l)
    
def fweval(expr, g=globals(), l=locals()):
    """
    Write expr, evaluate it and write result to fstream.
    
    Write algorithm:
        - write L{desceval} and expr in quotes.
        - evaluate expr using eval.
        - if exception occur then write L{descevalexc} and quoted exception.
        - if no error then write L{descevalrez} and quoted result.
        - write L{descevalln}.
        
    Then write this function you must get globals() and locals() if you using
    global or local scope variables into exprwssion.
    
    For example: C{ dbg.feval('dir(a)', globals(), locals()) }
    
    @param expr: Expression string.
    @type expr: str
    @param g: Global variables.
    @type g: dict = globals()
    @param l: Local variables.
    @type l: dict = locals()
    @return: Result of evaluation or Exception.
    @rtype: object or Exception
    """
    return _eval(fwriteln, expr, g, l)
    
    

#-------------------------------------------------------------------------------

# Register shutdown function.
atexit.register(shutdown)



#-DEBUG-START-------------------------------------------------------------------

def _debug():
    """
    Test function.
    
    Output to outstream is a next::
        
        *** DEBUG START ***
        
        >456
        ..>tst1:test
        >CALL tst2()
        ..>tst2 start
        ..>tst1:from tst2
        ..>CALL tst3()
        ....>tst3
        ..>RET from tst3()
        ..>tst2 end
        >RET from tst2()
        >CALL tst2()
        ..>tst2 start
        ..>tst1:from tst2
        ..>CALL tst3()
        ....>tst3
        ..>RET from tst3()
        ..>tst2 end
        >RET from tst2()
        arg1arg2>
        rwrite return: ('arg1', 'arg2')
        >EVAL: "str(5+6)"
        >  ER="11"
        > 
        >
        weval return: 11
        >CALL tst4()
        >RET from tst4()
        >Exception OK
        
        
        *** DEBUG END ***

    """
    def tst1(a):
        writeln('tst1:', a)
    
    @wlevel    
    def tst2():
        writeln('tst2 start')
        tst1('from tst2')
        tst3()
        writeln('tst2 end')
    
    @wlevel    
    def tst3():
        writeln('tst3')
        print tst3.func_name
    
    @wlevel
    def tst4():
        raise AttributeError
        
    writeln(456)
    fwriteln(456)
    lvlup()
    tst1('test')
    lvldown()
    tst2()
    lvldown()
    lvldown()
    lvldown()
    tst2()
    writeln('\nrwrite return: %s' % (rwrite('arg1', 'arg2'), ) )
    writeln('\nweval return: %s' % (weval('str(5+6)'), ) )
    try:
        tst4()
    except AttributeError:
        writeln('Exception OK')
    else:
        writeln('Exception fails')
    
    pass
    
#-DEBUG-END---------------------------------------------------------------------

if __name__=='__main__':
    _debug()

