#!/usr/bin/env python
# -*- coding: utf-8 -*-
#--------------------------------------------------------------------
# Filname: handler.py
# Author: Mazhugin Aleksey
# Created: 2007/11/05
# ID: $Id: handler.py 13 2008-11-29 18:31:31Z alex $
# URL: $URL: file:///myfiles/svn_/XPyLIB/trunc/XPyLIB/handler.py $
# Copyright: Copyright (c) 2007, Mazhugin Aleksey.
# License: BSD
#--------------------------------------------------------------------

"""
General handlers implementation.

Implement operation with collection of objects (handler-objects) same as a single
object.

Quik start.
===========

Program source::
    
    def _debug():
        import sys
        h1=sys.stdout
        h2=sys.stderr
        ho=Handler()
        ho.handler_add(h1,h2)
        ho.write('Test handlerE{\}n')

    
In this programm you add two handler-objects (I{stdout} and I{stderr})
in B{Handler} I{ho}. Then you call function from this handler-objects.
Detail of function calling:
    - In first you get new B{Handler} object with attributes I{write} used as
        handler-object I{ho.write}.
    - You do function call for each handler-object I{('Test handler\\n')}.
As rezult this action will be writen string I{'Test handler\\n'} into two stream.


@warning: For all index operation may return exception as L{list} object if
    argument is wrong.
"""

# Importing
import XPyLIB.dbg as dbg ##@


# Start Implementation

class HandlerError(Exception):
    """
    Represent error or exception then calling handler.
    """
    def __init__(self, exc):
        """
        @param exc: Exception instance.
        @type exc: Exception
        """
        self.exc=exc
        
    def __str__(self):
        """
        String convertation.
        """
        return self.__class__.__name__ + '(): ' + str(self.exc)

class Handler(object):
    """
    Base handler class.
    
    Workflow of handler processing is follow:
    - Call C{filter(event="start", ("action", None), None)}.
    - If returned C{True} enter handler loop (for each registered handler):
        - Call C{filter("pre", ("call", None), handler, *args, **kwds)} for "call",
            C{filter("pre", ("get", name), handler)} for "get",
            C{filter("pre", ("set", name), handler, value)} for set.
        - If returned C{True}:
            - Call C{handler.action(*args, **kwds)} for "call" or getattr and
                setattr for "get" and "set".
            - Call C{filter("post", ("action", None), handler, result)}.
    - Call C{filter("stop", ("action", None), None)}.
    
    Pre processing filter functon must return C{True} for processing current handler and
    C{False} for skip. Post processing filter must return C{True} for
    continue processing remain handlers or C{False} for stop processing.
    
    
    @ivar __handler_hl: Registered handlers list.
    @type __handler_hl: list of handler instances
    
    @ivar handler_filter: Filter function for handlers.
        Called each time befor work with each handler.
        Function must return C{True} for enable handler execution, or C{False} for
        disable. Function I{B{C{ fun (event, action, handler, *args, **kwds) }}} where:
            - B{event} - event type:
                - "I{start}" - Initialization, befor start handlers loop.
                - "I{pre}" - Pre handlers loop.
                - "I{post}" - Post handlers loop.
                - "I{stop}" - Deinitialization after handlers loop.
            - B{action} - Tuple of (function, name) of method requested from handler.
                - I{function} - "call", "get", "set" for calling, getattr and setattr.
                - I{name} - name for attribute for "get" and "set", and None for "call".
            - B{handler} - Handler instance.
            - B{args} - Arguments passed to handler.
            - B{kwds} - Keywords passed to handler.
            
            Additional arguments:
            - For "pre" + "call" - is function arguments.
            - For "pre" + "set" arg[0] - is value to setup.
            - For "post" arg[0] is result from handler.
    @type handler_filter: function or None
    
    """
    
    def __init__(self, filter = None):
        """
        @ivar handler_filter: Filter function for handlers.
        @type handler_filter: function or None
        """
        
        self.__handler_hl = []
        
        self.handler_filter = filter
        
    def handler_add(self, *handlers):
        """
        Add handler-object.
        
        @param handlers: Handler-objects for append.
            Add only new handlers (unique for this list).
            For not unique do nothing.
        @type handlers: handler-objects instance
        @return: Index of last added handler-object.
        @rtype: int
        """
        for handler in handlers:
            if handler not in self.__handler_hl:
                # Add only unique.
                self.__handler_hl.append(handler)
        return len(self.__handler_hl)-1
        
    def handler_insert(self, index, *handlers):
        """
        Insert handler-object.
        
        @param handlers: Handler-objects for inserting.
            Insert only new handlers (unique for this list).
            For not unique do nothing.
        @type handlers: handler-objects instance
        @return: Index of handler that point index befor insertion.
        @rtype: int
        """
        i=index
        for handler in handlers:
            if handler not in self.__handler_hl:
                # Add only unique.
                self.__handler_hl.insert(i, handler)
                i+=1
        return i
    
    def handler_remove(self, *handlers):
        """
        Remove handler-objects.
        
        @param handlers: Handler-objects for removing.
        @type handlers: handler-object instance
        """
        for handler in handlers:
            self.__handler_hl.remove(handler)
        
    def handler_pop(self, index):
        """
        Remove and return handler-objects by index.
        
        @param handlers: Handler-objects for removing.
        @type handlers: handler-object instance.
        @return: Poped handler.
        @rtype: handler-object
        """
        return self.__handler_hl.pop(index)
            
    def handler_has_handler(self, *handlers):
        """
        Test handler-objects for exists.
        
        @param handlers: Handler-objects for test.
        @type handlers: handler-object instance
        @return: List of handler-object's index if exist or B{-1} if not find.
        @rtype: int
        """
        r=[]
        for handler in handlers:
            try:
                i=self.__handler_hl.index(handler)
            except ValueError:
                i=-1
            r.append(i)
        return r
    
    def handler_getHandler(self, index):
        """
        Get handler-object by index.
        
        @param index: Handler's object index.
        @type index: int
        @return: Handler-bject.
        @rtype: Handler-object instance
        """
        return self.__handler_hl[index]
    
    def __len__(self):
        """
        Count handler-objects.
        
        @return: Handler-objects count.
        @rtype: int
        """
        return len(self.__handler_hl)
    
    def __call__(self, *args, **kwds):
        """
        Execute call for each handler-object I{ho(*args, **kwds)}.
        
        @return: List returned values from handler-objects. For error handler-objects
            return L{HandlerError} instance instead object's rezult. Filtered
            handlers is skiped from results. Error in filters is skiped and equivalent
            True.
        """
        rez=[]
        ##dbg.fwriteln('__call__ __handler_hl==', self.__handler_hl) ##@
        hl = self.__handler_hl
        # Start filter.
        c = False
        if len(hl):
            c = True # Continue condition.
            if self.handler_filter:
                try:
                    c = self.handler_filter("start", ("call", None), None)
                except:
                    pass
            ##if
        ##if
        # Handler filter.
        if c:
            for h in hl:
                c = True
                if self.handler_filter:
                    try:
                        c = self.handler_filter("pre", ("call", None), h, *args, **kwds)
                    except:
                        pass
                ##if
                if c:
                    try:
                        r=h(*args, **kwds)
                    except Exception, e:
                        r=HandlerError(e)
                    rez.append(r)
                    ##dbg.fwriteln('*** call ret ', r) ##@
                    if self.handler_filter:
                        try:
                            c = self.handler_filter("post", ("call", None), h, r)
                            ##dbg.fwriteln('*** call filter ', c) ##@
                        except Exception,e:
                            ##dbg.fwriteln('*** call filter EXC ', e) ##@
                            pass
                    if not c:
                        ##dbg.fwriteln('BREAK') ##@
                        break # Stop handler loop.
                ##if
            ##for
        ##if c
        # Stop filter.
        if len(hl):
            if self.handler_filter:
                try:
                    self.handler_filter("stop", ("call", None), None)
                except:
                    pass
            ##if
        ##if

        return rez
    
    def __getattr__(self, name):
        """
        Get new handler of attribute's values from handler-objects.
         
        @return: New of attribute's values from handler-objects.
        @rtype: Handler
        """
        ##dbg.fwriteln('__getattr__.name=', name) ##@
        ##bg.fwriteln('__getattr__. __dict__=', self.__dict__) ##@
        r=Handler()
        hl=self.__dict__['_Handler__handler_hl']
        hf=self.__dict__['handler_filter']
        c = False
        # Start filter.
        if len(hl):
            c = True # Continue condition.
            if hf:
                try:
                    c = hf("start", ("get", None), None)
                except:
                    pass
            ##if
        ##if
        
        # Handler filter.
        if c:
            for h in hl:
                c = True
                if hf:
                    try:
                        c = hf("pre", ("get", name), h)
                    except:
                        pass
                ##if
                if c:
                    try:
                        rr=getattr(h, name)
                    except AttributeError, e:
                        rr=HandlerError(e)
                    r.handler_add(rr)
                    if hf:
                        try:
                            c = hf("post", ("get", name), h, rr)
                        except:
                            pass
                    if not c:
                        break # Stop handler loop.
                ##if
            ##for
        ##if c
            
        # Stop filter.
        if len(hl):
            if hf:
                try:
                    hf("stop", ("get", None), None)
                except:
                    pass
            ##if
        ##if
        
        return r
    
    def __setattr__(self,name,value):
        """
        Set attrebute for all handler-objects.
        
        @warning: For self may set only "__handler_hl" attribute. 
            All other attributes try set to handler-objects.
        
        @return: List of rezults for handler-objects:
            None if no error, or Exception instance
            if error occur.
        @rtype: list
        """
        ##dbg.fwriteln('__setattr__. name=', name) ##@
        # Set instance attributes.
        if name in ('_Handler__handler_hl', 'handler_filter'):
            object.__setattr__(self, name, value)
            return
        # Set handler-objects attributes.
        rez=[]
        hl = self.__handler_hl
        c = False
        # Start filter.
        if len(hl):
            c = True # Continue condition.
            if self.handler_filter:
                try:
                    c = self.handler_filter("start", ("get", None), None)
                except:
                    pass
            ##if
        ##if
        
        # Handler filter.
        if c:
            for h in hl:
                c = True
                if self.handler_filter:
                    try:
                        c = self.handler_filter("pre", ("set", name), h, value)
                    except:
                        pass
                ##if
                if c:
                    try:
                        r=setattr(h, name, value)
                    except Exception, e:
                        r=HandlerError(e)
                    rez.append(r)
                    if self.handler_filter:
                        try:
                            c = self.handler_filter("post", ("set", name), h, r)
                        except:
                            pass
                    if not c:
                        break # Stop handler loop.
                ##if
            ##for
        ##if c
            
        # Stop filter.
        if len(hl):
            if self.handler_filter:
                try:
                    self.handler_filter("stop", ("set", None), None)
                except:
                    pass
            ##if
        ##if
                    
        return rez
        
    def __delattr__(self,name):
        """
        Delete attrebute for all handler-objects.
        """
        for h in self.__handler_hl:
            try:
                ##dbg.fwriteln('__del__: ', name) ##@
                delattr(h, name)
            except Exception, e:
                ##dbg.fwriteln('__del__ exception', e) ##@
                pass
        
    def __str__(self):
        """
        Return informal string.
        """
        ##dbg.fwriteln('__str__') ##@
        s='Handler instance: ['
        for h in self.__handler_hl:
            s+=str(h)+', '
        if self.__handler_hl:
            s=s[:-2]
        s+=']'
        return s
    
    
class EventFilter(object):
    """
    Event filter for L{Handler}.
    
    Emulate event handler (with L{Handler}). When has event and it buble from
    one object to another. If any object is return specified result 
    (True for example) then bubling is stoped.
    Filter is filtered only "call" action.
    See L{EventFilter.__init__} and L{EventFilter.__call__}.
    """
    
    def __init__(self, StopIfTrue = True):
        """
        @param StopIfTrue: Initialize L{stopIfTrue} variable.
        @type StopIfTrue: bool = True
        
        @ivar stopIfTrue: StopIfTrue: Stop handler loop condition - if value this parameter
            is equal result from handler then handler iterations will be stoped,
            otherwise continue. Filter only "call" action.
        @type stopIfTrue: bool
        """
        
        self.stopIfTrue = StopIfTrue
        
    def __call__(self, event, action, handler, *args, **kwds):
        """
        Callable function for handler filter.
        
        See L{Handler.filter}.
        
        @param event: Event type. May be "start", "pre", "post", "stop" events.
        @type event: str
        @param action: Tuple of action - ("action", "name"). action may be
            "call", "get", "set". name - is a requested attribute name for
            "get" and "set" and None for "call".
        @type action: tuple
        @param handler: Handler instance (not Handler class) for which called
            filter.
        @type handler: object
        @param args: Additional arguments.
            - For "pre" + "call" - is function arguments.
            - For "pre" + "set" arg[0] - is value to setup.
            - For "post" arg[0] is result from handler.
        @param kwds: Additional keywords. Applied only for "call" action.
        @return: True for continue loop or False otherwise.
        @rtype: bool
        """
        
        if event == "post" and action[0] == "call":
            if not (bool(args[0]) ^ self.stopIfTrue):
                # Stop handler loop if have StopIfTrue equal result.
                return False
        # Continue handlers loop otherwise.
        return True
    
    
# End Implemetation

#-DEBUG-START-------------------------------------------------------------------

def _debug():
    """
    """
    import sys
    h1=sys.stdout
    h2=sys.stderr
    ho=Handler()
    ho.handler_add(h1,h2)
    ho.write('Test handler\n')
    print str(ho)
    print str(ho.__class__)
    print str(ho.__dict__)
    
    
    del sys
    pass
    
#-DEBUG-END---------------------------------------------------------------------

if __name__=='__main__':
    _debug()
