#!/usr/bin/env python
# -*- coding: utf-8 -*-
#--------------------------------------------------------------------
# Filname: __init__.py
# Author: Mazhugin Aleksey
# Created: 2007/11/04
# ID: $Id: xlogging.py 13 2008-11-29 18:31:31Z alex $
# URL: $URL: file:///myfiles/svn_/XPyLIB/trunc/XPyLIB/xlogging.py $
# Copyright: Copyright (c) 2007, Mazhugin Aleksey.
# License: BSD
#--------------------------------------------------------------------
# Package initialisation

"""
Some extensions for Python logging package.

This module import standart Python logging namespace " from logging import * ",
and you may use only xlogging for most usual targets.

X{Logging config file format.}
==============================
    Logging file format like config format in python logging package, but
    have some differents.
    
    The file must contain I{main} sections called [loggers], [handlers],
    [filters] and  [formatters] which identify by name the entities of each
    type which are defined in the file. For each such entity, there is
    a separate I{entity} section which identified how that entity is configured.
    Thus, for a logger named log01 in the [loggers] section, the relevant
    configuration details are held in a section [logger_log01]. Similarly,
    a handler called hand01 in the [handlers] section will have its
    configuration held in a section called [handler_hand01], while a
    formatter called form01 in the [formatters] section will have its
    configuration specified in a section called [formatter_form01].
    The root logger configuration must be specified in a I{entity} section called
    [logger_root].
    The manager section configure loggers manager. Available only
    one option: "disable". (It as logging.disable() function). This
    section and option are optional.
    
    In I{main} sections you must use option B{keys} (it mandatory):

        - B{keys} - Add new object or will change configuration of existing
            object.
        
    Examples of these sections in the file are given below.::
        
        [manager]
        disable=WARNING
        
        [loggers]
        keys=root, log02, log03
        
        [handlers]
        keys=hand01, hand02, hand03, hand04
        
        [formatters]
        keys=form01,form02,form03,form04,form05,form06,form07,form08,form09
        
        [filters]
        keys=filt01, filt02, filt03
    
    The entity sections may contains B{actions} option. In this
    option may be listed action's names (it is a options's names with
    description of action). Each action entity have next format::
        
        act01=command, objecttype, arg01, arg02, ...
    
    where:
        - act01 - name of action given in actions option.
        - command - command that need to executed.
        - objecttype - object type for which command executed.
        - arg01, arg02, ... - additional arguments for command.
    
    For entity's sections B{logger_???} and B{handler_???} you may execute next
    commands: B{clear}, B{remove}, B{add}.
    For entity's sections B{formatter_???} and B{filter_???} actions will be
    I{ignored}.
    
    The I{arg**} in commands is name of objects.
    
    As I{objecttype} for sections you must use next types:
        - for B{[logger_???]} : handler, filter.
        - for B{[handler_???]} : filter, formatter.
    
    For entity section you also can specify object attribute and value to set::
        
        attributes=attr01, attr02
        attr01=attrname, attrvalue
        attr02=attrname, attrvalue
    
    Where B{attributes} is a option with attribute's options
    names I{attr01}, I{attr02}.
    Into options I{attr01}, I{attr02} you must specify attribute name
    B{attrname} and attribute value B{attrvalue}. The value willbe evaluated
    using eval() in logging namespace.
    
    The root logger I{must} specify a B{level}, list of handlers and filters.
    For [logger_???] sections have no mandatory options. By default
    level, handlers and other attributes are no changed.
    An example of a root logger section is given below.::
        
        [logger_root]
        level=NOTSET
        handlers=hand01
        filters=
        actions=act01, act02
        act01=clear, handler
        act02=remove, filter, filt99
        act03=add, handler, hand55
        disabled=0
        attributes=attr01
        attr01= name, A.B.C
    
    For loggers other than the root logger, some additional information is
    required.
    The level and handlers entries are interpreted as for the root logger,
    except that if a non-root logger's level is specified as NOTSET,
    the system consults loggers higher up the hierarchy to determine the
    effective level of the logger. The B{propagate} entry is set to B{1}
    (is default) to indicate
    that messages must propagate to handlers higher up the logger hierarchy
    from this logger, or B{0} to indicate that messages are not propagated to
    handlers up the hierarchy. The B{qualname} entry is the hierarchical channel
    name of the logger, that is to say the name used by the application to
    get the logger.
    The B{disabled} set this logger disabled (I{1}) or enabled (I{0}). By
    default disabled is I{0} for new or not changed for existing logger.
    The mandatory options is no. If B{qualname} is not specified then
    configuration name is used, for below example it is 'parser'.
    The B{filters} entry, specify filters names to use.If this entry specified
    then clear all filters then add new filters from I{filters} list.
    If I{handlers} entry specified then clear all filters then add new
    handlers from I{handlers} list.
    This is illustrated by the following example.::
        
        [logger_parser]
        level=DEBUG
        handlers=hand01
        propagate=1
        qualname=compiler.parser
        filters=
        actions=act01
        act01=clear,handler
        disabled=0
        attributes=attr01
        attr01= name, A.B.C
    
    Sections which specify handler configuration are exemplified by the following. 
    The B{class} entry indicates the handler's class (as determined by eval() in
    the logging package's namespace). The level is interpreted as for loggers,
    and NOTSET is taken to mean "log everything". 

    The B{formatter} entry indicates the key name of the formatter for
    this handler. If blank, a default formatter (C{logging._defaultFormatter})
    is used. If a name is specified, it must appear in the [formatters]
    section and have a corresponding section in the configuration file. 

    The B{args} entry, when eval() evaluated in the context of the logging
    package's namespace, is the list of arguments to the constructor
    for the handler class. Refer to the constructors for the relevant
    handlers, or to the examples below, to see how typical entries are
    constructed.
    
    The B{filters} entry, specify filters names to use.If this entry specified
    then clear all filters then add new filters from I{filters} list.
    
    The mandatory options is B{class} and B{args} and only for new handlers.::
        
        [handler_hand01]
        class=StreamHandler
        level=NOTSET
        formatter=form01
        args=(sys.stdout,)
        filters=
        actions=act01
        act01=add, filter, filt01
        attributes=attr01
        attr01= attrname, A.B.C
    
    Sections which specify filter configuration are typified by the following.
    The mandatory options is B{class} and B{args} and only for new filters.::
        
        [filter_filt01]
        class=Filter
        args=('App')
        attributes=attr01
        attr01= name, A.B.C
    
    Sections which specify formatter configuration are typified by the following.::
        
        [formatter_form01]
        format=F1 %(asctime)s %(levelname)s %(message)s
        datefmt=
        class=logging.Formatter
        args=()
        attributes=attr01
        attr01= name, A.B.C
    
    The mandatory options is B{format} and B{datefmt}.
    
    The B{format} entry is the overall format string, and the B{datefmt}
    entry is the strftime()-compatible date/time format string.
    If empty, the package substitutes ISO8601 format date/times,
    which is almost equivalent to specifying the date format string
    "The ISO8601 format also specifies milliseconds, which are appended to
    the result of using the above format string, with a comma separator.
    An example time in ISO8601 format is 2003-01-23 00:29:50,411. 
    
    The B{class} entry is optional. It indicates the name of the formatter's
    class (as a dotted module and class name.) This option is useful for
    instantiating a Formatter subclass. Subclasses of Formatter can present
    exception tracebacks in an expanded or condensed format. 
    
    


@warning: You must accurate use remove or clear options or commands,
    becouse may be conflict with currently executed code in runtime if it
    use removed objects.

@note: After importing xlogging L{XLogger} class set as default class for
new loggers.

@var TRACE: Logging level constant, used for prereliase tracing.
@type TRACE: int

@var DBGTRACE: Logging level constant, used for debug tracing.
@type DBGTRACE: int

@var _xBaseloggerClass: Base logger class (get from logging). Used as base for XLogger.
@type _xBaseloggerClass: Logger
"""

# Importing
import logging
from logging import *
from functools import wraps

try:
    import thread
    import threading
except:
    thread=None

# Not XPyLIB.dbg, only dbg for cross links avoid.
import dbg ##@

##dbg.fwriteln('*** xlogging debug output ***') ##@

# Constants.

TRACE = 15
DBGTRACE = 5

# Add logging level constants.
addLevelName(TRACE, 'TRACE')
addLevelName(DBGTRACE, 'DBGTRACE')

# Store base logger class.
_xBaseloggerClass = logging.getLoggerClass()

# Start Implementation
##@dbg.fwlevel ##@
def getLevelByName(levelname):
    """
    Return level name by level value.
    
    If no such level then raise error KeyError.
    """
    ##dbg.fwriteln('getLevelByName: level %s=%d'% (levelname,logging._levelNames[levelname])) ##@
    return logging._levelNames[levelname]
        
##@dbg.fwlevel ##@    
def applyConfigFile(fname, defaults=None):
    """
    Read the logging configuration from a ConfigParser-format file.

    This can be called several times from an application, allowing an end user
    the ability to select from various pre-canned configurations (if the
    developer provides a mechanism to present the choices and load the chosen
    configuration).
    In versions of ConfigParser which have the readfp method [typically
    shipped in 2.x versions of Python], you can pass in a file-like object
    rather than a filename, in which case the file-like object will be read
    using readfp.
    
    Algorithm:
    ==========
        
        1. Read [loggers], [handlers], [filters], [formatters] keys into DICT.
        2. FOR _logger_ IN [loggers]:
            1. IF _logger_ EXIST get it, ELSE create new
            2. Read _fs_=[filters].
            3. IF _fs_ THEN:
                1. _filters_.clear.
                2. FOR _f_ IN _fs_:
                    1. Read filter from config.
                    2. Create new filter.
                    3. Add filter to _logger_.
            4. FOR _f_ IN _fs_:
                1. Read filter changes from config section.
                2. Change filter _f_.
                3. Read [attributes] and apply.
            5. Read _hs_=[handlers].
            6. IF _hs_ THEN:
                1. _handlers_.clear.
                2. FOR _h_ IN _hs_:
                    1. Read handler from config.
                    2. Create new handler.
                    3. Add handler to _logger_.
            7. FOR _h_ IN _hs_:
                1. Read handler changes from config section.
                2. Change handler _h_.
                3. Read [formatter] and ADD or CHANGE.
                4. Read [filters] and ADD or CHANGE.
                5. Read [attributes] and apply.
            8. Read [attributes] and apply.
            
    Logger section example::
        
        [manager]
        #disable=WARNING
        
        [logger_parser]
        level=DEBUG
        handlers=hand01
        propagate=1
        qualname=compiler.parser
        filters=
        actions=act01
        act01=clear,handler
        disabled=0
        attributes=attr01
        attr01= name, A.B.C
    """
    import ConfigParser

    ##dbg.fwriteln('Load config parser.') ##@
    
    cp = ConfigParser.SafeConfigParser(defaults)
    if hasattr(cp, 'readfp') and hasattr(fname, 'readline'):
        cp.readfp(fname)
    else:
        cp.read(fname)
    
    # Test for mandatory main sections.
    ##dbg.fwriteln('Test for mandatory main sections.') ##@
    if not all( \
        [ cp.has_section(s) and cp.has_option(s, 'keys') \
        for s  in ('loggers', 'handlers', 'filters', 'formatters') ] ):
        ##dbg.fwriteln('\tFAIL Test for mandatory main sections.') ##@
        return # Return if not all have.
    
    loggers=dict((n.strip(), None) for n in str(cp.get('loggers','keys')).split(',') if n.strip())
    handlers=dict((n.strip(), None) for n in str(cp.get('handlers','keys')).split(',') if n.strip())
    filters=dict((n.strip(), None) for n in str(cp.get('filters','keys')).split(',') if n.strip())
    formatters=dict((n.strip(), None) for n in str(cp.get('formatters','keys')).split(',') if n.strip())
    
    ##dbg.fwriteln('loggers=', loggers) ##@
    ##dbg.fwriteln('handlers=', handlers) ##@
    ##dbg.fwriteln('filters=', filters) ##@
    ##dbg.fwriteln('formatters=', formatters) ##@
    
    # Start critical section.
    logging._acquireLock()
    try:
    
        ##dbg.fwriteln('manager configure') ##@
        sec='manager'
        if cp.has_section(sec):
            if cp.has_option(sec, 'disable'):
                try:
                    ##dbg.fwriteln('set disable=',cp.get(sec,'disable')) ##@
                    disable( getLevelByName(cp.get(sec,'disable')) )
                except:
                    ##dbg.fwriteln('\texception') ##@
                    pass
                    
        ##dbg.fwriteln('\nSTART LOGGER LOOP\n') ##@
        for logger in loggers:
            sec = 'logger_' + logger
            ##dbg.fwriteln('Read section=', sec) ##@
            if cp.has_section(sec):
                opts = _options(sec, cp)
                ##dbg.fwriteln('opts=', opts) ##@
                # Get logger name, overwork about root.
                if logger == 'root':
                    name = logger
                else:
                    name = opts.has_key('qualname') and opts['qualname'] or logger
                ##dbg.fwriteln('name=', name) ##@
                # Get logger, overwork about root logger.
                if name == 'root':
                    loginst = root
                else:
                    loginst = logging.getLogger(name)
                loggers[logger] = loginst
                # Configure logger.
                ##dbg.fwriteln('Configure logger') ##@
                if opts.has_key('level'):
                    try:
                        loginst.setLevel( getLevelByName(opts['level']) )
                        ##dbg.fwriteln('set level=', getLevelByName(opts['level'])) ##@
                    except:
                        ##dbg.fwriteln('\texception') ##@
                        pass
                if opts.has_key('propagate'):
                    try:
                        loginst.propagate = int(opts['propagate'])
                        ##dbg.fwriteln('set propogate', int(opts['propagate'])) ##@
                    except:
                        ##dbg.fwriteln('\texception') ##@
                        pass
                if opts.has_key('disabled'):
                    try:
                        loginst.disabled = int(opts['disabled'])
                        ##dbg.fwriteln('set disabled', int(opts['disabled'])) ##@
                    except:
                        ##dbg.fwriteln('\texception') ##@
                        pass
                    
                # Apply attributes.
                ##dbg.fwriteln('Apply attributes') ##@
                _applyAttrs(opts, loginst)
                
                # Apply filters.
                ##dbg.fwriteln('Apply filters') ##@
                if opts.has_key('filters'):
                    # Clear filters.
                    ##dbg.fwriteln('\tClearfilters') ##@
                    for f in loginst.filters:
                        loginst.remove(f)
                    fs=[ n.strip() for n in opts['filters'].split(',') if n.strip() ]
                    ##dbg.fwriteln('\tfilters=', fs) ##@
                    _applyFilters(fs, loginst, cp, filters)
                    
                # Apply handlers.
                ##dbg.fwriteln('Apply handlers') ##@
                if opts.has_key('handlers'):
                    # Clear filters.
                    ##dbg.fwriteln('\tClear handlers') ##@
                    for h in loginst.handlers:
                        loginst.removeHandler(h)
                    hs=[ n.strip() for n in opts['handlers'].split(',') if n.strip() ]
                    ##dbg.fwriteln('handlers=', hs) ##@
                    _applyHandlers(hs, loginst, cp, handlers, formatters, filters)
                        
                # Apply actions.
                ##dbg.fwriteln('Apply actions') ##@
                _applyActions(opts, loginst, cp, handlers, formatters, filters)
                
            ##end if cp.has_section(sec):
                   
        ##end for logger
        ##dbg.fwriteln() ##@

    finally:
        # End critical section.
        logging._releaseLock()
        
##@dbg.fwlevel ##@
def _applyAttrs(opts, inst):
    """
    Apply attributes from dictionary "opts" to object instance "inst".
    
    @param opts: Dictionary with item "attributes" and items listed in
        "attributes" which will be applied.
    @type opts: dict
    @param inst: Object instance.
    """
    if opts.has_key('attributes'):
        ##dbg.fwriteln('Find attributes') ##@
        for a in [ opts[n.strip()] for n in opts['attributes'].split(',') \
            if opts.has_key( n.strip() ) ]:
            # Loop for attributes.
            aa = [n.strip() for n in a.partition(',')]
            ##dbg.fwriteln('apply attr "',aa[0], '" = ', aa[2]) ##@
            if aa[0]:
                try:
                    ##dbg.fwriteln('\tset attr') ##@
                    setattr(inst, aa[0], aa[2])
                except:
                    ##dbg.fwriteln('\t\texception') ##@
                    pass
        ##end for a

##@dbg.fwlevel ##@ 
def _options(sec, cp, raw=0):
    """
    Return readed options and it's values in dictionary.
    
    @param sec: Section name.
    @type sec: str
    @param cp: Config parser instance.
    @type cp: ConfigParser
    @param raw: Raw mode for get in ConfigParser.
    @type raw: int = {0}
    @return: Dictionary with options and values.
    @rtype: dict
    """
    r = {}
    if cp.has_section(sec):
        r = dict( (n, cp.get(sec,n,raw)) for n in cp.options(sec) )
        ##dbg.fwriteln('options=', r) ##@
    return r

##@dbg.fwlevel ##@
def _applyFilters(flts, inst, cp, filters):
    """
    Aplly filters with names listed in "flts": to object "inst".
    
    First try find named filter into "inst.filters" list (which have attribute "name"),
    if not find then try find into "filters" dict,
    else create new. Then apply changes
    to filter.
    Filter name must exist in filters list.
    To add filter to inst used "addFilter" method.
    To created or added filters set attribute "name" to filter name.
    
    @param flts: List with filter's names.
    @type flts: list
    @param inst: Object instance to apply filters.
    @type inst: Object instance
    @param cp: Config parser instance.
    @type cp: ConfigParser
    @param filters: Permited filters name with existing or None instances.
    @type filters: dict
    
    Filter section example::
        
        [filter_filt01]
        class=Filter
        args=('App')
        attributes=attr01
        attr01= name, A.B.C
        
    """
    if not flts or not hasattr(inst, 'filters') or not hasattr(inst, 'addFilter'):
        return
    import sys
    ##_self_module=sys.modules[XLogger.__module__]
    ##dbg.fwriteln('_self_module=', _self_module) ##@
    ##dbg.fwriteln('Filter loop') ##@
    for fn in flts:
        ##dbg.fwriteln('fn=', fn) ##@
        if not filters.has_key(fn):
            # Exit for if no such filter name into filters.
            ##dbg.fwriteln('EXIT: no such name into filters') ##@
            break
        if cp.has_section('filter_' + fn):
            opts = _options('filter_' + fn, cp)
            ##dbg.fwriteln('opts=', opts) ##@
            f = None # Filter
            
            # Try find filter in instance.
            ##dbg.fwriteln('Try find filter in instance') ##@
            for fi in inst.filters:
                if hasattr(fi, 'name') and (fi.name == fn):
                    ##dbg.fwriteln('FIND filter in instance') ##@
                    f=fi
                    break
            if not f:
                if filters[fn] != None:
                    ##dbg.fwriteln('find filter into filters') ##@
                    f=filters[fn]
                else:
                    # Create new filter.
                    ##dbg.fwriteln('Create new filter') ##@
                    klass = logging.Filter
                    if opts.has_key('class'):
                        c = opts['class'].strip()
                        ##dbg.fwriteln('["class"]=', c) ##@
                        if c:
                            try:
                                ##dbg.fwriteln('eval class') ##@
                                c = eval(c, vars(sys.modules[__name__]))
                            except:
                                ##dbg.fwriteln('\texception') ##@
                                c=None
                        if c:
                            ##dbg.fwriteln('Set new class', c) ##@
                            klass = c
                    a = ()
                    ##dbg.fwriteln('search args') ##@
                    if opts.has_key('args'):
                        ##dbg.fwriteln('have args') ##@
                        a = opts['args']
                        ##dbg.fwriteln('args=', a) ##@
                        try:
                            ##dbg.fwriteln('Try eval args') ##@
                            a = eval(a, vars(sys.modules[__name__]))
                        except:
                            ##dbg.fwriteln('\texception') ##@
                            a=()
                            
                    ##dbg.fwriteln('CREATE filter') ##@    
                    f = klass(*a)
                ##end if filters[fn] != None
                try:
                    ##dbg.fwriteln('Add to instance') ##@
                    inst.addFilter(f)
                except:
                    pass
                
            ##end if not f
            if filters[fn] == None:
                ##dbg.fwriteln('set to filters') ##@
                filters[fn]=f
                
            # Set filter name.
            ##dbg.fwriteln('Set name to filter=',fn) ##@
            setattr(f, 'name', fn)
            
            ##dbg.fwriteln('Apply attrs') ##@
            _applyAttrs(opts, f)
            
        ##end if cp.has_section('filter_' + fn)
        ##dbg.fwriteln() ##@
    ##end for f
    ##dbg.fwriteln('End filter loop') ##@
    # Ignore actions for filters

##@dbg.fwlevel ##@
def _applyFormatter(fmtr, inst, cp, formatters):
    """
    Aplly formatter with name listed in "fmtr": to object "inst".
    
    First try find named formatter into "inst.formatter" list (which have attribute "name"),
    if not find then try find into "formatters" dict, 
    else create new. Then apply changes
    to formatter.
    Formatter name must exist in formatters list.
    To add filter to inst used "setFormatter" method.
    
    @warning: Options for formatter readed in raw mode (see ConfigParser).
    
    @param fmtr: Formatter name.
    @type fmtr: list
    @param inst: Object instance to apply formatter.
    @type inst: Object instance
    @param cp: Config parser instance.
    @type cp: ConfigParser
    @param formatters: Permited formatters name with existing or None instances.
    @type formatters: dict
    
    Formatter section example::
        
        [formatter_form01]
        format=F1 %(asctime)s %(levelname)s %(message)s
        datefmt=
        class=logging.Formatter
        args=()
        attributes=attr01
        attr01= name, A.B.C
            
    """
    if not fmtr or not hasattr(inst, 'formatter') or not hasattr(inst, 'setFormatter'):
        return
    import sys
    ##_self_module=sys.modules[XLogger.__module__]
    ##dbg.fwriteln('_self_module=', _self_module) ##@
    sec = 'formatter_' + fmtr
    ##dbg.fwriteln('sec=', sec) ##@
    if cp.has_section(sec):
        opts = _options(sec, cp, 1) # Raw
        ##dbg.fwriteln('opts=', opts) ##@
        f = None # Formatter
        if not formatters.has_key(fmtr):
            # Exit if no such formatter name into formatterss.
            ##dbg.fwriteln('EXIT such formatter is not in formatters.') ##@
            return
         # Try find formatter in instance.
        ##dbg.fwriteln('Try find in instance') ##@
        fi = inst.formatter
        if hasattr(fi, 'name') and (fi.name == fmtr):
            ##dbg.fwriteln('FIND in instance') ##@
            f=fi
        if not f:
            ##dbg.fwriteln('search in formatters') ##@
            if formatters[fmtr] != None:
                ##dbg.fwriteln('FIND in formatters') ##@
                f=formatters[fmtr]
            else:
                # Create new formatter.
                ##dbg.fwriteln('Create new') ##@
                klass = logging.Formatter
                if opts.has_key('class'):
                    c = opts['class'].strip()
                    ##dbg.fwriteln('["class"]=',c) ##@
                    if c:
                        try:
                            ##dbg.fwriteln('Eval class') ##@
                            c = eval(c, vars(sys.modules[__name__]))
                        except:
                            ##dbg.fwriteln('\texception') ##@
                            c=None
                    if c:
                        ##dbg.fwriteln('Set new class ', c) ##@
                        klass = c
                a = ()
                ##dbg.fwriteln('Search args') ##@
                if opts.has_key('args'):
                    a = opts['args']
                    ##dbg.fwriteln('args=', a) ##@
                    try:
                        ##dbg.fwriteln('eval args') ##@
                        a = eval(a, vars(sys.modules[__name__]))
                    except:
                        ##dbg.fwriteln('\texception') ##@
                        a=()
                if opts.has_key('format'):
                    fs = opts['format']
                    ##dbg.fwriteln('format=', fs) ##@
                else:
                    ##dbg.fwriteln('no format') ##@
                    fs = None
                if opts.has_key('datefmt'):
                    dfs = opts['datefmt']
                    ##dbg.fwriteln('datefmt=', dfs) ##@
                else:
                    ##dbg.fwriteln('no datefmt') ##@
                    dfs = None
                ##dbg.fwriteln('Create new formatter.') ##@
                f = klass(fs, dfs, *a)
            ##end if formatters[fmtr] != None
            try:
                ##dbg.fwriteln('Set instance formatter.') ##@
                inst.setFormatter(f)
            except:
                ##dbg.fwriteln('\texception') ##@
                pass
        ##end if not f
        if formatters[fmtr] == None:
            ##dbg.fwriteln('Set formatters') ##@
            formatters[fmtr]=f
        ##dbg.fwriteln('Set name to formatter=', fmtr) ##@
        setattr(f,'name',fmtr)
        
        ##dbg.fwriteln('Apply attrs') ##@
        _applyAttrs(opts, f)
    ##end if cp.has_section('filter_' + fn)
    # Ignore actions for formatter

##@dbg.fwlevel ##@
def _applyHandlers(hdls, inst, cp, handlers, formatters, filters):
    """
    Aplly handlerss with names listed in "hdls": to object "inst".
    
    First try find named handler into  "inst.handlers" list
    (which have attribute "name"), if not then
    try find in  "handlers" dict,
    else create new. Then apply changes
    to handler.
    Handler name must exist in handlers list.
    To add handler to inst used "addHandler" method.
    
    @param hdls: List with hamdler's names.
    @type hdls: list
    @param inst: Object instance to apply handlers.
    @type inst: Object instance
    @param cp: Config parser instance.
    @type cp: ConfigParser
    @param handlers: Permited handlers name with existing or None instances.
    @type handlers: dict
    @param formatters: Permited formatters.
    @type formatters: dict
    @param filters: Permited filters.
    @type filters: dict
    
    Handler section example::
        
        [handler_hand01]
        class=StreamHandler
        level=NOTSET
        formatter=form01
        args=(sys.stdout,)
        filters=
        actions=act01
        act01=add, filter, filt01
        attributes=attr01
        attr01= attrname, A.B.C
        
    """
    if not hdls or not hasattr(inst, 'handlers') or not hasattr(inst, 'addHandler'):
        return
    import sys
    ##_self_module=sys.modules[XLogger.__module__]
    ##dbg.fwriteln('_self_module=', _self_module) ##@
    ##dbg.fwriteln('HANDLERS LOOP') ##@
    for hn in hdls:
        ##dbg.fwriteln('hn=', hn) ##@
        if cp.has_section('handler_' + hn):
            opts = _options('handler_' + hn, cp)
            ##dbg.fwriteln('opts=', opts) ##@
            h = None # Handler
            if not handlers.has_key(hn):
                # Exit if no such handler name into handlers.
                ##dbg.fwriteln('EXIT such handler not find in handlers') ##@
                return
            # Try find handler in instance.
            ##dbg.fwriteln('Try find in instance') ##@
            for hi in inst.handlers:
                if hasattr(hi, 'name') and (hi.name == hn):
                    ##dbg.fwriteln('FIND in instance') ##@
                    h=hi
                    break
            if not h:
                # Try find handlers dict.
                ##dbg.fwriteln('Search in handlers') ##@
                if handlers[hn] != None:
                    ##dbg.fwriteln('FIND in handlers') ##@
                    h=handlers[hn]
                else:
                    # Create new handler.
                    ##dbg.fwriteln('Create new') ##@
                    klass = NullHandler
                    if opts.has_key('class'):
                        c = opts['class'].strip()
                        ##dbg.fwriteln('["class"]=', c) ##@
                        if c:
                            try:
                                ##dbg.fwriteln('Eval class') ##@
                                c = eval(c, vars(sys.modules[__name__]))
                            except Exception, e:
                                ##dbg.fwriteln('\texception=', e) ##@
                                c=None
                        if c:
                            ##dbg.fwriteln('Set new class') ##@
                            klass = c
                    a = ()
                    if opts.has_key('args'):
                        a = opts['args']
                        ##dbg.fwriteln('args=', a) ##@
                        try:
                            ##dbg.fwriteln('Eval args') ##@
                            a = eval(a, vars(sys.modules[__name__]))
                        except:
                            ##dbg.fwriteln('\texception') ##@
                            a=()
                    
                    ##dbg.fwriteln('Create and set new class', klass) ##@
                    h = klass(*a)
                ##end if handlers[hn] != None
                ##dbg.fwriteln('Try add to instance') ##@
                try:
                    ##dbg.fwriteln('Add handler to instance') ##@
                    inst.addHandler(h)
                except:
                    ##dbg.fwriteln('\texception') ##@
                    pass
                
            ##end if not h    
            if handlers[hn] == None:
                ##dbg.fwriteln('Add to handlers') ##@
                handlers[hn]=h
            
            ##dbg.fwriteln('Set name to handler=',hn) ##@
            setattr(h, 'name', hn)
            
            # Configure handler.
            ##dbg.fwriteln('Configure handler') ##@
            if opts.has_key('level'):
                try:
                    h.setLevel( getLevelByName(opts['level']) )
                    ##dbg.fwriteln('set level=', getLevelByName(opts['level'])) ##@
                except:
                    ##dbg.fwriteln('\texception') ##@
                    pass
                    
            # Apply formatter.
            if opts.has_key('formatter'):
                ##dbg.fwriteln('Apply formatter') ##@
                _applyFormatter(opts['formatter'].strip(), h, cp, formatters)
            # Apply filters.
            if opts.has_key('filters'):
                ##dbg.fwriteln('ApplyFilters') ##@
                _applyFilters([f.strip() for f in opts['filters'].split(',') if f.strip()], h, cp, filters)
            
            ##dbg.fwriteln('apply attrs') ##@
            _applyAttrs(opts, h)
            
            ##dbg.fwriteln('Apply actions') ##@
            _applyActions(opts, h, cp, handlers, formatters, filters)
        ##end if cp.has_section('handler_' + hn)
        ##dbg.fwriteln() ##@
    ##end for h

##@dbg.fwlevel ##@
def _applyActions(opts, inst, cp, handlers, formatters, filters):
    """
    Aplly actions with names listed in "opts['actions']": to object "inst".
    
    First try find named filter into "filters" dict, if not find then
    (which have attribute "name") into "inst.filters" list,
    else create new. Then apply changes
    to filter.
    Filter name must exist in filters list.
    To add filter to inst used "addFilter" method.
    
    @param opts: Dictionary with actions describes.
    @type opts: dict
    @param inst: Object instance to apply actions.
    @type inst: Object instance
    
    Action example::
        
        actions=act01
        act01=add, filter, filt01
    
    """

    if opts.has_key('actions'):
        ##dbg.fwriteln('Action LOOP') ##@
        for a in [ opts[n.strip()] for n in opts['actions'].split(',') \
            if opts.has_key( n.strip() ) ]:
            # Loop for actions.
            aa = [n.strip() for n in a.split(',')]
            ##dbg.fwriteln('aa=', aa) ##@
            if aa[0] == 'add':
                ##dbg.fwriteln('add') ##@
                if aa[1] == 'filter' and ( issubclass(inst.__class__, logging.Handler) \
                    or issubclass(inst.__class__, logging.Logger) ):
                    ##dbg.fwriteln('\tfilters') ##@
                    _applyFilters([aa[2]], inst, cp, filters)
                elif aa[1] == 'handler' and issubclass(inst.__class__, logging.Logger):
                    ##dbg.fwriteln('\thandler') ##@
                    _applyHandlers([aa[2]], inst, cp, handlers, formatters, filters)
                elif aa[1] == 'formatter' and issubclass(inst.__class__, logging.Handler):
                    ##dbg.fwriteln('\tformatter') ##@
                    _applyFormatter(aa[2], inst, cp, formatters)
            ##end elif a[0] == 'add'
            elif aa[0] == 'remove':
                ##dbg.fwriteln('remove') ##@
                if aa[1] == 'filter' and ( issubclass(inst.__class__, logging.Handler) \
                    or issubclass(inst.__class__, logging.Logger) ):
                    ##dbg.fwriteln('\tfilters') ##@
                    for i in inst.filters:
                        if hasattr(i, 'name') and i.name == aa[2]:
                            inst.removeFilter(i)
                            break
                elif aa[1] == 'handler' and issubclass(inst.__class__, logging.Logger):
                    ##dbg.fwriteln('\thandlers') ##@
                    for i in inst.handlers:
                        if hasattr(i, 'name') and i.name == aa[2]:
                            inst.removeHandler(i)
                            break
                elif aa[1] == 'formatter' and issubclass(inst.__class__, logging.Handler):
                    if hasattr(inst.formatter, 'name') and inst.formatter.name == aa[2]:
                        ##dbg.fwriteln('formatter') ##@
                        inst.setFormatter(logging.Formatter())
            ##end elif a[0] == 'remove'
            elif aa[0] == 'clear':
                ##dbg.fwriteln('clear') ##@
                if aa[1] == 'filter' and ( issubclass(inst.__class__, logging.Handler) \
                    or issubclass(inst.__class__, logging.Logger) ):
                    ##dbg.fwriteln('\tfilters') ##@
                    for i in inst.filters:
                        inst.removeFilter(i)
                elif aa[1] == 'handler' and issubclass(inst.__class__, logging.Logger):
                    ##dbg.fwriteln('\thandlers') ##@
                    for i in inst.handlers:
                        inst.removeHandler(i)
                elif aa[1] == 'formatter' and issubclass(inst.__class__, logging.Handler):
                    ##dbg.fwriteln('\tformatter') ##@
                    inst.setFormatter(logging.Formatter())
            ##end elif a[0] == 'clear'
            ##dbg.fwriteln() ##@
        ##end for a


class NullHandler(Handler):
    """
    Handler that do no output.
    """
    def __init__(self, level=NOTSET):
        ##dbg.fwriteln('NullHandler.__init__') ##@
        Handler.__init__(self, level)
        
    def emit(self, record):
        """
        Do nothing.
        """
        ##dbg.fwriteln('NullHandler.emit') ##@
        pass

class RangeFilter:
        """
        Filter available output only if level in specified range.
        Return true if available.
        """
        def __init__(self, min=NOTSET, max=CRITICAL ):
            """
            @param min: Minimal available level for output.
            @type min: int = {NOTSET}
            @param max: Maximal available level for output.
            @type max: int = {CRITICAL}
            """
            ##dbg.fwriteln('RangeFilter.__init__() ', min, ', ', max) ##@
            self._min=min
            self._max=max
            
        def filter(self, record):
            """
            Return true if message is available (in specified range).
            """
            ##dbg.fwriteln('RangeFilter.filter %d <= %d <= %d == %d' % (self._min,record.levelno, self._max,  self._min <= record.levelno <= self._max) ) ##@
            return self._min <= record.levelno <= self._max

class NamesFilter:
        """
        Filter available output only if logger name equal any from 'names'.
        Return true if available.
        """
        def __init__(self, names='', case=0 ):
            """
            @param names: List (or string with one name) of names with which
            allow output.
            @type names: list or str = ''
            @param case: If true then compare names case sensitive.
            @type case: bool = 0
            """
            ##dbg.fwriteln('NamesFilter.__init__() names=', repr(names)) ##@
            if type(names) == str:
                names = [names]
            if not case:
                self._names=[n.lower() for n in names]
            else:
                self._names=names
            self._case=case
            
        def filter(self, record):
            """
            Return true if message is available (in 'names').
            """
            r= self._case and record.name or record.name.lower()
            ##dbg.writeln('**** name=',r) ##@
            for n in self._names:
                ##dbg.writeln('******** n=',n) ##@
                if r.startswith(n):
                    return 1
            ##dbg.writeln('**** return 0') ##@
            return 0
        
class IndentFormatter(Formatter, object):
    """
    Formatter with indentation level support.
    
    Instance of this class exists only in one exemplar. For other requests
    of new instance returned already existing. 
    
    Level indentation is specific for each threads separately (if thread
    library available, otherwase used only common indentation level for all threads).
    
    @cvar _level: Indentation level. Dictionary of {thread.id, level} if thread
        enebled, else level.
    @type _level: dict{int,int} or int
    @cvar __instance: Singular instance.
    @type __instance: Instance IndentFormatter
    """
    
    __instance = None
    _level = thread and threading.local() or 0
    
    def __new__(cls, *args):
        """
        Allow only one exemplar of instance. For other calls return existing.
        """
        ##dbg.fwriteln('IndentFormatter.__new__') ##@
        logging._acquireLock()
        try:
            if not IndentFormatter.__instance:
                ##dbg.fwriteln('IndentFormatter.__new__ create new') ##@
                IndentFormatter.__instance = super(IndentFormatter, cls).__new__(cls, *args)
                IndentFormatter.__createLock()
        finally:
            logging._releaseLock()
        return IndentFormatter.__instance
    
    def __init__(self, fmt=None, datefmt=None):
        """
        Initialize the formatter with specified format strings.

        Initialize the formatter either with the specified format string, or a
        default as described above. Allow for specialized date formatting with
        the optional datefmt argument (if omitted, you get the ISO8601 format).
        """
        ##dbg.fwriteln('IndentFormatter.__init__') ##@
        Formatter.__init__(self, fmt, datefmt)
    
    @staticmethod
    def __init_level():
        """
        Initialize level for current thread.
        """
        ##dbg.fwriteln('IndentFormatter.initlevel') ##@
        if thread:
            # Init thread dependent levels.
            if not hasattr(IndentFormatter._level, 'level'):
                ##dbg.fwriteln('IndentFormatter.initlevel initialized') ##@
                IndentFormatter._level.level = 0
    
    @staticmethod
    def __get_level():
        """
        Get level for current thread.
        """
        ##dbg.fwriteln('IndentFormatter.getlevel') ##@
        IndentFormatter.__init_level()
        if thread:
            l=IndentFormatter._level.level
        else:
            l=IndentFormatter._level
        return l
    
    @staticmethod
    def __set_level(value):
        """
        Set level for current thread.
        """
        ##dbg.fwriteln('IndentFormatter.setlevel') ##@
        IndentFormatter.__init_level()
        if thread:
            IndentFormatter._level.level=value
        else:
            IndentFormatter._level=value
    
    @staticmethod 
    def __createLock():
        """
        Acquire a thread lock for serializing access to the level counter.
        """
        ##dbg.fwriteln('IndentFormatter.createLock') ##@
        if thread:
            IndentFormatter.__lock = threading.RLock()
        else:
            IndentFormatter.__lock = None
    
    @staticmethod
    def acquire():
        """
        Acquire thread lock.
        """
        ##dbg.fwriteln('IndentFormatter.acquire') ##@
        if IndentFormatter.__lock:
            IndentFormatter.__lock.acquire()

    @staticmethod
    def release():
        """
        Release thread lock.
        """
        ##dbg.fwriteln('IndentFormatter.release') ##@
        if IndentFormatter.__lock:
            IndentFormatter.__lock.release()
    
    def format(self, record):
        """
        Format the specified record as text.
        """
        ##dbg.fwriteln('IndentFormatter.format') ##@
        c=__import__('XPyLIB')
        IndentFormatter.acquire()
        try:
            r=(c.log_levelbodychar * c.log_levelwidth * IndentFormatter.__get_level()) + c.log_levelendchar \
                + Formatter.format(self, record)
        finally:
            IndentFormatter.release()
        return r

    @staticmethod          
    def lvlup():
        """
        Increase indentation level.
        """
        ##dbg.fwriteln('IndentFormatter.lvlup') ##@
        IndentFormatter.acquire()
        try:
            IndentFormatter.__set_level(IndentFormatter.__get_level()+1)
        finally:
            IndentFormatter.release()

    @staticmethod      
    def lvldown():
        """
        Decrease indentation level.
        """
        ##dbg.fwriteln('IndentFormatter.lvldown') ##@
        IndentFormatter.acquire()
        try:
            IndentFormatter.__set_level(IndentFormatter.__get_level()-1)
            if IndentFormatter.__get_level() < 0:
                IndentFormatter.__set_level(0)
        finally:
            IndentFormatter.release()

    @staticmethod              
    def lvlreset():
        """
        Reset indentation level to "0".
        """
        ##dbg.fwriteln('IndentFormatter.lvlreset') ##@
        IndentFormatter.acquire()
        try:
            IndentFormatter.__set_level(0)
        finally:
            IndentFormatter.release()

    @staticmethod        
    def lvlwrap(fun):
        """
        Auto level decorator for functions.
        
        Call L{lvlup}() befor function call and L{lvldown}() after 
        wrapped function return.
        @param fun: Wrapped function.
        @type func
        """
        ##dbg.fwriteln('IndentFormatter.lvlwrap') ##@
        @wraps(fun)
        def wrapper(*args, **kwds):
            IndentFormatter.acquire()
            try:
                l=IndentFormatter.__get_level()
                IndentFormatter.lvlup()
                r = fun(*args, **kwds)
                ##IndentFormatter.lvldown()
                IndentFormatter.__set_level(l) # Defense from error
            finally:
                IndentFormatter.release()
            return r
        return wrapper
    
    @staticmethod        
    def lvlwraplog(fun, loginst):
        """
        Auto level decorator for functions.
        
        Write first 'Call "fun.__name__"' using loginst.L{dbgtrace}().
        Also as L{lvlwrap}().
        If exception occurs then raise it.
        If level error tested then write '*** DBG LEVEL ERROR = different'
        where 'different' is level different.
        Write last 'Return from "fun.__name__"' using loginst.L{dbgtrace}().
        
        @param fun: Wrapped function.
        @type func
        @param loginst: Logger instance to use for reporting.
        @type loginst
        """
        ##dbg.fwriteln('IndentFormatter.lvlwraplog') ##@
        @wraps(fun)
        def wrapper(*args, **kwds):
            ##dbg.fwriteln('IndentFormatter.lvlwraplog.wrapper') ##@
            IndentFormatter.acquire()
            c=__import__('XPyLIB')
            try:
                l=IndentFormatter.__get_level()
                loginst.dbgtrace('%s%s()' % (c.log_descfuncall,fun.__name__))
                IndentFormatter.lvlup()
                r = fun(*args, **kwds)
                IndentFormatter.lvldown()
                if IndentFormatter.__get_level()!=l:
                    IndentFormatter.__set_level(l) # Defense from error
                    loginst.dbgtrace('%s%d' % (c.log_levelerr, d))
            except Exception,e:
                IndentFormatter.__set_level(l)
                ##dbg.fwriteln('IndentFormatter.lvlwraplog.wrapper except ',e) ##@
                raise e
            finally:
                IndentFormatter.release()
            loginst.dbgtrace('%s%s()' % (c.log_descfunret,fun.__name__))
            return r
        return wrapper

# Initialize instance of Indent formatter.
IndentFormatter()

class XLogger(_xBaseloggerClass):
    def __init__(self, name, level=NOTSET):
        """
        Base logging class fo xlogging.
        
        After importing xlogging this class set as default class for new loggers.
        """
        _xBaseloggerClass.__init__(self, name, level)
    
    def trace(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'TRACE'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.trace("Houston, we have a %s", "thorny problem", exc_info=1)
        """
        ##dbg.fwriteln('XLogger.trace') ##@
        _trace(self, msg, *args, **kwargs)
            
    def dbgtrace(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'DBGTRACE'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.dbgtrace("Houston, we have a %s", "thorny problem", exc_info=1)
        """
        ##dbg.fwriteln('XLogger.dbgtrace') ##@
        _dbgtrace(self, msg, *args, **kwargs)
        
           
# Set XLogger as base logger class.
logging.setLoggerClass(XLogger)

# Additional module assigments.
##dbg.fwriteln('_self_module=', sys.modules[__name__]) ##@

# Additional fuctions for bound to logger class.
##@dbg.fwlevel ##@
def _trace(self, msg, *args, **kwargs):
    """
    Log 'msg % args' with severity 'TRACE'.

    To pass exception information, use the keyword argument exc_info with
    a true value, e.g.

    logger.trace("Houston, we have a %s", "thorny problem", exc_info=1)
    """
    ##dbg.fwriteln('_trace') ##@
    if self.manager.disable >= TRACE:
        return
    ##dbg.fwriteln('_trace ... %d>=%d'%(TRACE,self.getEffectiveLevel())) ##@
    if TRACE >= self.getEffectiveLevel():
        apply(self._log, (TRACE, msg, args), kwargs)

##@dbg.fwlevel ##@       
def _dbgtrace(self, msg, *args, **kwargs):
    """
    Log 'msg % args' with severity 'DBGTRACE'.

    To pass exception information, use the keyword argument exc_info with
    a true value, e.g.

    logger.dbgtrace("Houston, we have a %s", "thorny problem", exc_info=1)
    """
    ##dbg.fwriteln('_dbgtrace') ##@
    if self.manager.disable >= DBGTRACE:
        return
    ##dbg.fwriteln('_dbgtrace ... %d>=%d'%(DBGTRACE,self.getEffectiveLevel())) ##@
    if DBGTRACE >= self.getEffectiveLevel():
        apply(self._log, (DBGTRACE, msg, args), kwargs)
            
# Add if need trace and dbgtrace method to root logger.
if not hasattr(root, 'tarce'):
    setattr(root, 'trace', XLogger.trace)
    
if not hasattr(root, 'dbgtarce'):
    setattr(root, 'dbgtrace', XLogger.dbgtrace)

##@dbg.fwlevel ##@   
def trace(msg, *args, **kwargs):
    """
    Log a message with severity 'TRACE' on the root logger.
    """
    if len(root.handlers) == 0:
        basicConfig()
    apply(_trace, (root, msg,)+args, kwargs)

##@dbg.fwlevel ##@
def dbgtrace(msg, *args, **kwargs):
    """
    Log a message with severity 'DBGTRACE' on the root logger.
    """
    if len(root.handlers) == 0:
        basicConfig()
    apply(_dbgtrace, (root, msg,)+args, kwargs)
    



# End Implemetation




#-DEBUG-START-------------------------------------------------------------------

def _debug():
    """
    """
    import XPyLIB
    import os
    
    class _TestHandler(Handler):
        """
        Test handler that add at start message 'TestHandler emit:' and output it to
        sys.stderr.
        """
        def __init__(self, level=NOTSET):
            dbg.fwriteln('TestHandler.__init__') ##@
            Handler.__init__(self, level)
            
        def emit(self, record):
            """
            
            """
            dbg.fwriteln('TestHandler.emit') ##@
            import sys
            msg=self.format(record)
            sys.stderr.write('TestHandler emit: %s\n' % msg)
            pass
    
    class _TestFilter(Filter):
        """
        Test filter that available only if msg starts with 'yes'
        """
        attr='NOT SET ATTR'
        def __init__(self, args='NO ARGS'):
            dbg.fwriteln('TestFilter.__init__() ', args) ##@
            Filter.__init__(self, '') # No name
            
        def filter(self, record):
            """
            """
            dbg.fwriteln('TestFilter.filter msg="'+record.getMessage()[:5]+'..."') ##@
            dbg.fwriteln('TestFilter.filter attr=', self.attr) ##@
            return 'yes' == record.getMessage()[:3]
            pass
        
    _debug.TestHandler=_TestHandler
    _debug.TestFilter=_TestFilter
    #root.setLevel(TRACE) do this throw .cfg.
    p=os.path.join(XPyLIB.dir_root,'XPyLIB','test','xlogging1.cfg')
    dbg.writeln('p=',p)
    applyConfigFile(p)
    A=getLogger('A')
    trace('test trace msg')
    dbgtrace('test dbgtrace msg MUST NOT print')
    A.trace('test A.trace msg')
    A.warning('test A.warning msg')
    AB=getLogger('A.B')
    AB.error('test A.B.error msg')
    AB.dbgtrace('!!!!!!test A.B.dbgtrace msg must NOT print')
    ABC=getLogger('A.B.C')
    ABC.dbgtrace('yes!!!!!!test A.B.C.dbgtrace msg must NOT print')
    ABC.trace('!!!!!!test A.B.C.trace msg must NOT print')
    ABC.trace('yes test A.B.C.trace msg')
    
    p=os.path.join(XPyLIB.dir_root,'XPyLIB','test','xlogging2.cfg')
    dbg.writeln('p=',p)
    applyConfigFile(p)
    warning('test root warning msg')
    A.warning('test A warning msg')
    critical('test root critical msg')
    A.critical('test A critical msg')
    
    B=getLogger('B')
    B.propagate=0
    B.setLevel(TRACE)
    hB=StreamHandler(sys.stderr)
    B.addHandler(hB)
    BF=getLogger('B.F')
    BG=getLogger('B.g')
    BH=getLogger('B.H')
    fFG=NamesFilter(['B.F','B.G'])
    #fFG=Filter('B.F')
    hB.addFilter(fFG)
    BF.trace('BF.trace')
    BH.trace('BH.trace MUST NOT OUT')
    BG.trace('BG.trace')
    pass
    
#-DEBUG-END---------------------------------------------------------------------

if __name__=='__main__':
    _debug()
